@@ -19,6 +19,7 @@
 
 #include "seed-private.h"
 #include <dlfcn.h>
+#include <pthread.h>
 
 JSClassRef gobject_class;
 JSClassRef gobject_method_class;
@@ -41,6 +42,9 @@ seed_toggle_ref (gpointer data, GObject * object, gboolean is_last_ref)
 {
   JSValueRef wrapper;
 
+  if (!g_object_get_data (object, "js-ref"))
+    return;
+
   wrapper = (JSValueRef) data;
 
   if (is_last_ref)
@@ -96,8 +100,8 @@ seed_wrap_object (JSContextRef ctx, GObject * object)
   if (user_data)
     return user_data;
 
-  if (seed_next_gobject_wrapper)
-    js_ref = seed_next_gobject_wrapper;
+  if (pthread_getspecific(seed_next_gobject_wrapper_key))
+    js_ref = pthread_getspecific(seed_next_gobject_wrapper_key);
   else
     js_ref = seed_make_wrapper_for_type (ctx, type);
 
@@ -111,7 +115,7 @@ seed_wrap_object (JSContextRef ctx, GObject * object)
 
   seed_add_signals_to_object (ctx, js_ref, object);
 
-  seed_next_gobject_wrapper = NULL;
+  pthread_setspecific(seed_next_gobject_wrapper_key, NULL);
 
   return js_ref;
 }
@@ -446,8 +450,8 @@ gboolean
 seed_gi_make_argument (JSContextRef ctx,
 		       JSValueRef value,
 		       GITypeInfo * type_info,
-		       GIArgInfo * arg_info,
-		       GArgument * arg, JSValueRef * exception)
+		       GArgument * arg,
+		       JSValueRef * exception)
 {
   GITypeTag gi_tag = g_type_info_get_tag (type_info);
 
@@ -485,12 +489,16 @@ seed_gi_make_argument (JSContextRef ctx,
       arg->v_uint32 = seed_value_to_uint (ctx, value, exception);
       break;
     case GI_TYPE_TAG_LONG:
+      arg->v_long = seed_value_to_long (ctx, value, exception);
+      break;
     case GI_TYPE_TAG_INT64:
-      arg->v_int64 = seed_value_to_long (ctx, value, exception);
+      arg->v_int64 = seed_value_to_int64 (ctx, value, exception);
       break;
     case GI_TYPE_TAG_ULONG:
+      arg->v_ulong = seed_value_to_ulong (ctx, value, exception);
+      break;
     case GI_TYPE_TAG_UINT64:
-      arg->v_uint64 = seed_value_to_ulong (ctx, value, exception);
+      arg->v_uint64 = seed_value_to_uint64 (ctx, value, exception);
       break;
     case GI_TYPE_TAG_INT:
       arg->v_int = seed_value_to_int (ctx, value, exception);
@@ -499,8 +507,10 @@ seed_gi_make_argument (JSContextRef ctx,
       arg->v_uint = seed_value_to_uint (ctx, value, exception);
       break;
     case GI_TYPE_TAG_SIZE:
+      arg->v_size = seed_value_to_size (ctx, value, exception);
+      break;
     case GI_TYPE_TAG_SSIZE:
-      arg->v_int = seed_value_to_int (ctx, value, exception);
+      arg->v_ssize = seed_value_to_ssize (ctx, value, exception);
       break;
     case GI_TYPE_TAG_FLOAT:
       arg->v_float = seed_value_to_float (ctx, value, exception);
@@ -532,8 +542,8 @@ seed_gi_make_argument (JSContextRef ctx,
 
 	arg->v_pointer = NULL;
 
-	if (interface_type == GI_INFO_TYPE_OBJECT
-	    || interface_type == GI_INFO_TYPE_INTERFACE)
+	if (interface_type == GI_INFO_TYPE_OBJECT ||
+	    interface_type == GI_INFO_TYPE_INTERFACE)
 	  {
 	    gobject = seed_value_to_object (ctx, value, exception);
 	    required_gtype =
@@ -574,10 +584,13 @@ seed_gi_make_argument (JSContextRef ctx,
 	    g_base_info_unref (interface);
 	    break;
 	  }
-	else if (interface_type == GI_INFO_TYPE_STRUCT)
+	else if (interface_type == GI_INFO_TYPE_STRUCT ||
+		 interface_type == GI_INFO_TYPE_UNION)
 	  {
 	    if (JSValueIsObjectOfClass (ctx, value, seed_struct_class))
 	      arg->v_pointer = seed_pointer_get_pointer (ctx, value);
+	    else if (JSValueIsObjectOfClass (ctx, value, seed_union_class))
+	      arg->v_pointer = seed_pointer_get_pointer (ctx, value);
 	    else
 	      {
 		GType type =
@@ -678,7 +691,6 @@ seed_gi_make_argument (JSContextRef ctx,
 	      {
 		SeedNativeClosure *privates = seed_make_native_closure (ctx,
 									(GICallableInfo *) interface,
-									arg_info,
 									value);
 		arg->v_pointer = privates->closure;
 		g_base_info_unref (interface);
@@ -759,18 +771,21 @@ seed_gi_argument_make_js (JSContextRef ctx,
     case GI_TYPE_TAG_UINT32:
       return seed_value_from_uint (ctx, arg->v_uint32, exception);
     case GI_TYPE_TAG_LONG:
+      return seed_value_from_long (ctx, arg->v_long, exception);
     case GI_TYPE_TAG_INT64:
-      return seed_value_from_long (ctx, arg->v_int64, exception);
+      return seed_value_from_int64 (ctx, arg->v_int64, exception);
     case GI_TYPE_TAG_ULONG:
+      return seed_value_from_ulong (ctx, arg->v_ulong, exception);
     case GI_TYPE_TAG_UINT64:
-      return seed_value_from_ulong (ctx, arg->v_uint64, exception);
+      return seed_value_from_uint64 (ctx, arg->v_uint64, exception);
     case GI_TYPE_TAG_INT:
       return seed_value_from_int (ctx, arg->v_int32, exception);
     case GI_TYPE_TAG_UINT:
       return seed_value_from_uint (ctx, arg->v_uint32, exception);
     case GI_TYPE_TAG_SSIZE:
+      return seed_value_from_ssize (ctx, arg->v_ssize, exception);
     case GI_TYPE_TAG_SIZE:
-      return seed_value_from_int (ctx, arg->v_int, exception);
+      return seed_value_from_size (ctx, arg->v_size, exception);
     case GI_TYPE_TAG_FLOAT:
       return seed_value_from_float (ctx, arg->v_float, exception);
     case GI_TYPE_TAG_DOUBLE:
@@ -1040,7 +1055,7 @@ seed_gvalue_from_seed_value (JSContextRef ctx,
 
       if (o == NULL || g_type_is_a (G_OBJECT_TYPE (o), type))
 	{
-	  g_value_init (ret, G_TYPE_OBJECT);
+	  g_value_init (ret, type);
 	  g_value_set_object (ret, o);
 
 	  return TRUE;
@@ -1928,6 +1943,98 @@ seed_value_from_double (JSContextRef ctx, gdouble val, JSValueRef * exception)
   return JSValueMakeNumber (ctx, (gdouble) val);
 }
 
+/**
+ * seed_value_to_size:
+ * @ctx: A #SeedContext.
+ * @val: The #SeedValue to convert.
+ * @exception: A reference to a #SeedValue in which to store any exceptions.
+ *             Pass %NULL to ignore exceptions.
+ *
+ * Converts the given #SeedValue into a #gsize.
+ *
+ * Return value: The #gsize represented by @val, or %NULL if an exception
+ *               is raised during the conversion.
+ *
+ */
+gsize
+seed_value_to_size (JSContextRef ctx, JSValueRef val, JSValueRef * exception)
+{
+  if (!JSValueIsNumber (ctx, val) && !JSValueIsBoolean (ctx, val))
+    {
+      if (!JSValueIsNull (ctx, val))
+	seed_make_exception (ctx, exception, "ConversionError",
+			     "Can not convert Javascript value to" " gsize");
+      return 0;
+    }
+
+  return (gsize) JSValueToNumber (ctx, val, NULL);
+}
+
+/**
+ * seed_value_from_size:
+ * @ctx: A #SeedContext.
+ * @val: The #gsize to represent.
+ * @exception: A reference to a #SeedValue in which to store any exceptions.
+ *             Pass %NULL to ignore exceptions.
+ *
+ * Converts the given #gsize into a #SeedValue.
+ *
+ * Return value: A #SeedValue which represents @val, or %NULL if an exception
+ *               is raised during the conversion.
+ *
+ */
+JSValueRef
+seed_value_from_size (JSContextRef ctx, gsize val, JSValueRef * exception)
+{
+  return JSValueMakeNumber (ctx, (gdouble) val);
+}
+
+/**
+ * seed_value_to_ssize:
+ * @ctx: A #SeedContext.
+ * @val: The #SeedValue to convert.
+ * @exception: A reference to a #SeedValue in which to store any exceptions.
+ *             Pass %NULL to ignore exceptions.
+ *
+ * Converts the given #SeedValue into a #gssize.
+ *
+ * Return value: The #gssize represented by @val, or %NULL if an exception
+ *               is raised during the conversion.
+ *
+ */
+gssize
+seed_value_to_ssize (JSContextRef ctx, JSValueRef val, JSValueRef * exception)
+{
+  if (!JSValueIsNumber (ctx, val) && !JSValueIsBoolean (ctx, val))
+    {
+      if (!JSValueIsNull (ctx, val))
+	seed_make_exception (ctx, exception, "ConversionError",
+			     "Can not convert Javascript value to" " gssize");
+      return 0;
+    }
+
+  return (gssize) JSValueToNumber (ctx, val, NULL);
+}
+
+/**
+ * seed_value_from_ssize:
+ * @ctx: A #SeedContext.
+ * @val: The #gssize to represent.
+ * @exception: A reference to a #SeedValue in which to store any exceptions.
+ *             Pass %NULL to ignore exceptions.
+ *
+ * Converts the given #gssize into a #SeedValue.
+ *
+ * Return value: A #SeedValue which represents @val, or %NULL if an exception
+ *               is raised during the conversion.
+ *
+ */
+JSValueRef
+seed_value_from_ssize (JSContextRef ctx, gssize val, JSValueRef * exception)
+{
+  return JSValueMakeNumber (ctx, (gdouble) val);
+}
+
 /**
  * seed_value_to_string:
  * @ctx: A #SeedContext.
@@ -2021,11 +2128,16 @@ JSValueRef
 seed_value_from_string (JSContextRef ctx,
 			const gchar * val, JSValueRef * exception)
 {
-  JSStringRef jsstr = JSStringCreateWithUTF8CString (val);
-  JSValueRef valstr = JSValueMakeString (ctx, jsstr);
-  JSStringRelease (jsstr);
+  if (val == NULL)
+    return JSValueMakeNull (ctx);
+  else
+    {
+      JSStringRef jsstr = JSStringCreateWithUTF8CString (val);
+      JSValueRef valstr = JSValueMakeString (ctx, jsstr);
+      JSStringRelease (jsstr);
 
-  return valstr;
+      return valstr;
+    }
 }
 
 /**
@@ -2113,22 +2225,26 @@ seed_value_from_filename (JSContextRef ctx,
 {
   GError *e = NULL;
   gchar *utf8;
-
-  utf8 = g_filename_to_utf8 (val, -1, NULL, NULL, &e);
-
-  if (e)
+  
+  if (val == NULL)
+    return JSValueMakeNull (ctx);
+  else
     {
-      seed_make_exception_from_gerror (ctx, exception, e);
-      g_error_free (e);
-      // TODO: FIXMEShould be JS Null maybe?
-      return NULL;
+      utf8 = g_filename_to_utf8 (val, -1, NULL, NULL, &e);
+
+      if (e)
+        {
+          seed_make_exception_from_gerror (ctx, exception, e);
+          g_error_free (e);
+          return JSValueMakeNull (ctx);
+        }
+
+      JSValueRef valstr = seed_value_from_string (ctx, utf8, exception);
+
+      g_free (utf8);
+
+      return valstr;
     }
-
-  JSValueRef valstr = seed_value_from_string (ctx, utf8, exception);
-
-  g_free (utf8);
-
-  return valstr;
 }
 
 /**
@@ -2202,7 +2318,7 @@ seed_validate_enum (GIEnumInfo * info, long val)
 {
   gint n, i;
   GIValueInfo *value_info;
-  glong value;
+  gint value; // TODO: investigate glong/gint mismatch w/ g_value_info_get_value
 
   n = g_enum_info_get_n_values (info);
   for (i = 0; i < n; i++)
